(*fastDummyprinter, Copyright (C) 2012 Laure Gonnord*)
(*This file was part of Aspic Distribution, (C) Laure Gonnord 2007*)
(*  *)
(* This library is free software; you can redistribute it and/or *)
(* modify it under the terms of the GNU Lesser General Public *)
(* License as published by the Free Software Foundation; either *)
(* version 2.1 of the License, or (at your option) any later version. *)
(*  *)
(* This library is distributed in the hope that it will be useful, *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)
(*  *)
(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)


open Symbol
open Format

(*ocaml data types for expressions*)

type opnum = Plus | Times | Div | Minus

type expr_num = | OpNum of opnum * expr_num * expr_num
		| Ref of Symbol.t
		| CNat of int
		| CRat of Rational.t
		| Indet of Symbol.t

type opcompare = Eq | NotEq | SupEq | LessEq | Less | Sup
    
type term = ExprNum of opcompare *  expr_num * expr_num 
	    | Stateis of Symbol.t | CBool of bool | VBool of Symbol.t
		
type opboolB = And | Or | Imply 

type expr_bool = Term of term 
		 | OpBoolB of opboolB * expr_bool * expr_bool
		 | Neg of expr_bool

let debug = true


(*functions*)

let neg_opcompare = function 
  | Eq -> NotEq
  | NotEq -> Eq
  | SupEq -> Less
  | LessEq ->Sup
  | Less -> SupEq
  | Sup -> LessEq

let opcompare2string = function 
  | Eq ->  "="
  | NotEq -> "!="
  | SupEq -> ">="
  | LessEq -> "<="
  | Less ->  "<"
  | Sup ->  ">"

let opboolB2string = function 
  | And ->  "&&"
  | Or -> "||"
  | Imply -> "=>"


(*Multiply and simplify*)
let mult_basique x1 x2 = match (x1,x2) with
  | Ref(s1), _  ->  OpNum(Times,x1,x2)
  | _ , Ref(s2) ->  OpNum(Times,x1,x2)
  | CNat(1),e -> e
  | e,CNat(1) -> e
  | CNat(n1),CNat(n2) -> CNat(n1*n2)
  | CRat(r1),e when Rational.is_one r1 -> e
  | e,CRat(r2) when Rational.is_one r2 -> e
  | CRat(r1),CRat(r2) -> CRat(Rational.mul r1 r2)
  | CRat(r),CNat(n) -> 
      let r2 = Rational.mul_scal_int n r in
      let (b,i) = Rational.is_int r2 in 
	if b then CNat(i) else CRat(r2)
  | CNat(n),CRat(r) -> 
      let r2 = Rational.mul_scal_int n r in
      let (b,i) = Rational.is_int r2 in 
	if b then CNat(i) else CRat(r2)
  | _ -> OpNum(Times,x1,x2)


exception NonAffine

let is_indet (e:expr_num) = match e with
  | Indet(t) -> true
  | _ -> false

let is_id (id:Symbol.t) (e:expr_num)  = match e with 
  | Ref(t) -> (id=t)
  | _ -> false

(*returns true iff a given num expression is affine*)
let  is_affine (e1 : expr_num) = 
  let rec compte_ref exp = match exp with
    | Indet(s) -> raise NonAffine
    | Ref(s) -> (1,true)
    | CNat(n) -> (0,true)
    | CRat(rat) -> (0,true)
    | OpNum(op,e1,e2) -> 
	let (s1,b1) = compte_ref e1 and (s2,b2) =  compte_ref e2 in
	  match op with 
	    | Times -> if s1*s2 > 0 
	      then raise NonAffine else (s1+s2,true)
	    | _ -> (s1+s2,true)
  in try 
      (
	let (t,b) = compte_ref e1 in 
	  true
      )
    with
      | NonAffine -> false


(*distribute cte over a list of expressions*)
(*n or r >0 and l = [3,2x,7y,2w]*)
let rec ldistribue cst lexp =
  match cst with
      CNat(n)->
	List.map (fun el -> match el with
	      | Ref(w) -> OpNum(Times,CNat(n),Ref(w))
	      | CNat(u) -> CNat(n*u)
	      | CRat(r) -> 
		  let nr = Rational.mul_scal_int n r in
		  let (b,i) = Rational.is_int nr in 
		    if b then CNat(i) else CRat(nr)
  	      | OpNum(Times,CNat(u),Ref(w)) -> 
		  OpNum(Times,CNat(n*u),Ref(w)) 
	      | OpNum(Times,Ref(w),CNat(u)) -> 
		  OpNum(Times,CNat(n*u),Ref(w))
  	      | OpNum(Times,CRat(r),Ref(w)) -> 
		  let nr = Rational.mul_scal_int n r in
		  let (b,i) = Rational.is_int nr in 
		    if b then OpNum(Times,CNat(i),Ref(w))
		    else OpNum(Times,CRat(nr),Ref(w))
	      | OpNum(Times,Ref(w),CRat(r)) -> 
		  let nr = Rational.mul_scal_int n r in
		  let (b,i) = Rational.is_int nr in 
		    if b then OpNum(Times,CNat(i),Ref(w))
		    else OpNum(Times,CRat(nr),Ref(w))
		 ) lexp 
    | CRat(rr) -> 
	List.map (fun el -> match el with
		    | Ref(w) -> OpNum(Times,CRat(rr),Ref(w))
		    | CNat(u) -> 
			let nr = Rational.mul_scal_int u rr in
			let (b,i) = Rational.is_int nr in 
			  if b then CNat(i) else CRat(nr)
		    | CRat(r) -> 
			let nr = Rational.mul rr r in
			let (b,i) = Rational.is_int nr in 
			  if b then CNat(i) else CRat(nr)
  		    | OpNum(Times,CNat(u),Ref(w)) -> 
			let nr = Rational.mul_scal_int u rr in
			let (b,i) = Rational.is_int nr in 
			  if b then OpNum(Times,CNat(i),Ref(w))
			  else OpNum(Times,CRat(nr),Ref(w))
		    | OpNum(Times,Ref(w),CNat(u)) -> 
			let nr = Rational.mul_scal_int u rr in
			let (b,i) = Rational.is_int nr in 
			  if b then OpNum(Times,CNat(i),Ref(w))
			  else OpNum(Times,CRat(nr),Ref(w))
  		    | OpNum(Times,CRat(r),Ref(w)) -> 
			let nr = Rational.mul rr r in
			let (b,i) = Rational.is_int nr in 
			  if b then OpNum(Times,CNat(i),Ref(w))
			  else OpNum(Times,CRat(nr),Ref(w))
		    | OpNum(Times,Ref(w),CRat(r)) -> 
			let nr = Rational.mul rr r in
			let (b,i) = Rational.is_int nr in 
			  if b then OpNum(Times,CNat(i),Ref(w))
			  else OpNum(Times,CRat(nr),Ref(w))
		 ) lexp 
    | Ref(u) -> 
	List.map (fun el -> match el with
		    | Ref(w) -> OpNum(Times,Ref(u),Ref(w))
		    | CNat(n) -> OpNum(Times,CNat(n),Ref(u))
		    | CRat(r) -> OpNum(Times,CRat(r),Ref(u))
		    | _ -> failwith "distribue : not yet implemented"
		 ) lexp
    | OpNum(Times,cst2,Ref(u)) -> 
	let lres = ldistribue cst2 lexp in 
	  List.map (fun el -> match el with
		      | Ref(w) -> OpNum(Times,Ref(u),Ref(w))
		      | CNat(n) -> OpNum(Times,CNat(n),Ref(u))
		      | CRat(r) -> OpNum(Times,CRat(r),Ref(u))
		      | _ -> failwith "distribue :: not yet implemented"
		   ) lres
    | _ -> failwith "lmulscal : souci "
	
let list2exp lexp = 
  let resu = ref (CNat(0)) in
    List.iter ( fun one -> 
		  resu:=OpNum(Plus,one,!resu)
	      ) lexp ; !resu

      
(*l1.1 *(l2.1+l2.2+...) + l1.2 * (idem)*)
let develop_two_positive_lists l1 l2 = 
  let resu = ref [] in
    List.iter ( 
      fun oneel1 -> 
	resu:=(ldistribue oneel1 l2)@(!resu)
    ) l1 ; !resu

let developpe_ou_pas e ee = 
  let exp2 = mult_basique ee ee in
    (*pas de devl pour le moment*)
    exp2


let take_opposite exp = match exp with
    Ref(w) -> OpNum(Minus,CNat(0),Ref(w))
  | CNat(n) -> CNat(-n)
  | CRat(r) -> CRat(Rational.inv r)
  | _ -> failwith "take_opposite : pb"
  

(*For the rest of the functions, expressions are supposed to be affine*)
let rec pos_and_neg exp = 
  if not (is_affine exp) then failwith "pos and neg : should be affine";
(*   else fprintf std_formatter "affine exp : ok"; *)
  match exp with
    |  OpNum(op,e1,e2) as e -> (
	let (l1,r1) = pos_and_neg e1 
	and (l2,r2) = pos_and_neg e2 in
(* 	  fprintf std_formatter "ici : ok"; *)
	  match op with
	    | Minus -> 	(l1@r2,r1@l2)
	    | Plus -> (l1@l2,r1@r2)
	    | Times ->  (*ici on doit developper*)
		let left1 = develop_two_positive_lists l1 l2 and
		    left2 = develop_two_positive_lists r1 r2 and
		    right1 = develop_two_positive_lists r1 l2 and
		    right2 = develop_two_positive_lists r2 l1 in
		  (left1@left2,right1@right2)
	    | Div -> failwith "posandneg : div not yet implemented")
    | Ref(s) as e -> ([e],[])
    | CNat(n) as e -> if n>0 then ([e],[]) else ([],[e])
    | CRat(rat) as e -> if Rational.is_positive rat then  ([e],[]) else ([],[e])
    | Indet(t) -> failwith "pos and neg : indet should not happen"

(*throw away el in list el, returns true if el has been deleted*)
let rec enleve_el el l = match l with 
  | [] -> (l,false)
  | a::reste -> if a=el then (reste,true)
    else  let resu,b = enleve_el el reste in
      (a::resu,b)

(*thow away from lists elements that appear more than once*)
let rec enleve_doublons l1 l2 = match (l1,l2) with 
    | [], l2 -> (l1,l2)
    | l1, [] -> (l1,l2)
    | a::reste,l2 -> 
	let (l,b) = enleve_el a l2
	in if b then enleve_doublons reste l
	  else 
	    let la,lb = enleve_doublons reste l2
	    in (a::la,lb)

(*insert the element in list only if it doesn't exists*)
let rec insere el l = match l with
  | [] -> [el]
  | a::reste -> if a=el then l else a::(insere el reste)

let rec mergelists l1 l2 = 
  let resu = ref l2 in List.iter (fun el-> resu:= insere el !resu) l1;
    !resu


(*find num variables in numexpr*)
let rec find_lvars_num numexpr = match numexpr with
  | OpNum(_,e1,e2) ->( let l2 = find_lvars_num e2 and l1 = find_lvars_num e1 in
			 mergelists l1 l2)
  | Ref(t) -> [t]
  | CNat(n) -> []
  | CRat(r) -> []
  | Indet(t) -> [t]
      
      
(*find num variables in boolexpr*)
let rec find_lvars_bool boolexpr = match boolexpr with 
  | Term(t)-> (match t with 
		 | Stateis(w) -> []
		 | CBool(b) -> []
		 | VBool(v) -> [v]
		 | ExprNum(op,en1,en2) -> let l1 = find_lvars_num en1 and 
		       l2 =find_lvars_num en2 in mergelists l1 l2
	      )
  | OpBoolB(_,b1,b2)-> mergelists(find_lvars_bool b1) (find_lvars_bool b2)
  | Neg(b2) -> find_lvars_bool b2
	
let is_zero_rat = function
  | CRat(r) -> Rational.is_zero r
  | _ -> false

let rec print_list_with_sep fmt print_one_el l1 (s:string) = 
  let l1b = List.filter (
    fun el-> (el<>CNat(0) && not(is_zero_rat el))
  ) l1 in 
    match l1b with 
      | [] -> ()
      | [a] -> print_one_el fmt a 
      | a::reste -> (
	  fprintf fmt " %a %s" print_one_el a s;
	  print_list_with_sep fmt print_one_el reste s;
	)

let is_intconst  = function 
  | CNat(n) -> true
  | _ -> false

let get_nat = function 
  | CNat(n) -> n
  | _ -> failwith "getnat : is not a natural"

let is_ratconst  = function 
  | CRat(r) -> true
  | _ -> false

let get_rat = function 
  | CRat(r) -> r
  | _ -> failwith "getrat : is not a rational"

(*** auxiliary functions to simplify numerical expressions***)

(*constante de left-right*)
let collapse_constantes left right = 
  let pluscons (CNat(a)) (CNat(b)) = CNat(a+b) and
      pluscons2 (CRat(r1)) (CRat(r2)) = CRat(Rational.add r1 r2) in

  let collapse_constantes_aux lexp   = 
    let (ctesint,autres) = List.partition 
      (fun el -> is_intconst el) lexp in
    let resu1 = List.fold_left pluscons (CNat(0)) ctesint in 
    let (ctesrat,autres2) = List.partition 
      (fun el -> is_ratconst el) autres in
    let resu2 = List.fold_left pluscons2 (CRat(Rational.make 0 1)) 
      ctesrat in
      (resu1,resu2,autres2)
  in
  let (resu1a,resu2a,autres2a) = collapse_constantes_aux left 
  and (resu1b,resu2b,autres2b) = collapse_constantes_aux right in
  let nr = Rational.make ((get_nat resu1a) - (get_nat resu1b)) 1 
  and r = Rational.sub (get_rat(resu2a))  (get_rat(resu2b)) in
  let sum = Rational.add nr r in
  let (b,nn) = Rational.is_int sum in 
    if b then CNat(nn) else CRat(sum)
      
(*cherche la var s dans la liste et clacule les coeffs ds la somme *)
let collapse_variable lexp (s:Symbol.t) = 
  let cptnat = ref 0 and cptrat = ref (Rational.make 0 1) in
    List.iter
    (
      fun el -> match el with 
	| OpNum(Times,CNat(n),Ref(t)) when t=s -> cptnat := n+ !cptnat
	| OpNum(Times,Ref(t),CNat(n)) when t=s -> cptnat := n+ !cptnat
	| OpNum(Times,CRat(n),Ref(t)) when t=s -> 
	    cptrat := Rational.add n !cptrat
	| OpNum(Times,Ref(t),CRat(n)) when t=s -> 
	    cptrat := Rational.add n !cptrat
	| Ref(t) when t=s -> cptnat := 1 + !cptnat
	| _ -> () 
    ) lexp;
    let (b,n) = Rational.is_int !cptrat in
      if b then (cptnat:=n+ !cptnat; cptrat:=Rational.make 0 1);
      (!cptnat,!cptrat)
  

let collapse_vars l1 l2 lvars = 
  let pos = ref [] and neg = ref [] in 
  List.iter ( fun onevar ->
    let (n1,r1) = collapse_variable l1 onevar
    and (n2,r2) = collapse_variable l2 onevar in
    let nr = Rational.make (n1 - n2) 1 and r = Rational.sub r1 r2 in 
    let sum = Rational.add nr r in
    let (b,nn) = Rational.is_int sum in 
      if b then 
	begin 
	  if nn>0 
	  then pos := OpNum(Times,CNat(nn),Ref(onevar))::!pos
	  else 
	    if nn<0 then neg :=  OpNum(Times,CNat(-nn),Ref(onevar))::!neg
	end
      else
	if not(Rational.is_zero sum) then begin
	  if Rational.is_positive sum 
	  then pos := OpNum(Times,CRat(sum),Ref(onevar))::!pos
	  else 
	    neg := OpNum(Times,CRat(Rational.neg sum),Ref(onevar))::!neg
	end;
	    ) 
    lvars;
    (!pos,!neg)

let simplify_l l1 l2 lvars = 
  let (l1c,l2c) = collapse_vars l1 l2 lvars in
  let toto = collapse_constantes l1 l2 in
    match toto with 
      | CNat(n) as u -> if n>0 then (u::l1c,l2c) else if n<0 then (l1c,CNat(-n)::l2c)
	else (l1c,l2c)
      | CRat(r) as u -> if Rational.is_zero r then (l1c,l2c) else 
	  if Rational.is_positive r then  
	    (u::l1c,l2c) else (l1c,CRat(Rational.neg r)::l2c)

(**************print a numerical  expression******************)
let rec print_times fmt e ee lvars =
  let m = mult_basique e ee in
    match m with 	
      | OpNum(Times,CNat(n),x2) -> if n=1 then 
	  fprintf fmt "%a" (pp_print_expr_num lvars) x2  
	else fprintf fmt "%d*%a" n (pp_print_expr_num lvars) x2
      | OpNum(Times,x2,CNat(n)) -> if n=1 then 
	  fprintf fmt "%a" (pp_print_expr_num lvars) x2
	else fprintf fmt "%d*%a" n (pp_print_expr_num lvars) x2
      | OpNum(Times,CRat(r),x2) -> if Rational.is_one r then 
	  fprintf fmt "%a"  (pp_print_expr_num lvars) x2
	else fprintf fmt "%a*%a" Rational.print r 
	  (pp_print_expr_num lvars) x2
      | OpNum(Times,x2,CRat(r)) -> if Rational.is_one r then 
	  fprintf fmt "%a" ( pp_print_expr_num lvars) x2
	else fprintf fmt "%a*%a" Rational.print r 
	  (pp_print_expr_num lvars) x2
      | OpNum(Times,x1,x2)-> (
	  printf "je suis là " ;
	  let exp2 = developpe_ou_pas x1 x2  in 
	    if not (is_affine exp2) then failwith "sous exp non affine " 
	    else fprintf fmt "(%a) * (%a)"  (pp_print_expr_num lvars )x1
	      (pp_print_expr_num lvars) x2 
	)
	| _ ->  pp_print_expr_num lvars fmt  m
	  
and pp_print_expr_num lvars fmt (e1:expr_num)   = 
  match e1 with 
    | Ref(w) -> fprintf fmt "%s" (Symbol.to_string w)
    | CNat(n) -> fprintf fmt "%d" n
    | CRat(rat) -> Rational.print fmt rat
    | OpNum(Times,e,ee) -> print_times fmt e ee lvars
    | Indet(s) -> fprintf fmt "?" 
    | _ -> (let (l1,l2) = pos_and_neg e1 in
	    let (l1b,l2b) = simplify_l l1 l2 lvars in 
	    let t1 = List.length l1b and t2 = List.length l2b in 
	      print_list_with_sep fmt (pp_print_expr_num lvars) l1b " + ";
	      if t2 <> 0 then fprintf fmt " - ";  
	      print_list_with_sep fmt (pp_print_expr_num lvars) l2b " - ";
	      if t1+t2 = 0 then fprintf fmt "0";
	   )

(*Simplification of numerical expressions*)
let rec simplify_expr_num lvars e1 = match e1 with
    | Ref(w) as t -> t
    | CNat(n) as t -> t
    | CRat(rat) as t -> t
    | Indet(s) as t -> t
    | _ -> (
	let (l1,l2) = pos_and_neg e1 in
	let resu = ref (CNat(0)) in
	let (l1b,l2b) = simplify_l l1 l2 lvars in 
	let t1 = List.length l1b and t2 = List.length l2b in 
	  if t1+t2 = 0 then CNat(0) else
	    begin
	      if t1>0 then begin
		let tmp = List.hd l1b in 
		let resu2 = List.fold_right ( fun a b -> OpNum(Plus,a,b) )
		  (List.tl l1b) tmp in resu:= resu2; 
	      end;
	      if t2>0 then begin
		let tmp = List.hd l2b in 
		let resu2 = List.fold_right ( fun a b -> OpNum(Minus,a,b) )
		  (List.tl l2b) (take_opposite tmp) in resu:= resu2; 
	      end;
	      !resu;
	    end)

		  
(*Evaluate simple expressions, like 7==0 or 7>=0*)
let evaluate op exp  = 
  match (op,exp) with 
    | Eq,CNat(n) -> n=0 
    | Eq,CRat(r) -> Rational.is_zero r
    | NotEq,CNat(n) -> n<>0 
    | NotEq,CRat(r) -> not(Rational.is_zero r)
    | SupEq,CNat(n)  -> n>=0 
    | LessEq,CNat(n) -> n<=0
    | Less,CNat(n) -> n<0
    | Sup,CNat(n) ->  n>0
    | SupEq,CRat(r)  -> Rational.is_positive r
    | LessEq,CRat(r) -> Rational.is_zero r || not(Rational.is_positive r)
    | Less,CRat(r) -> not(Rational.is_positive r)
    | Sup,CRat(r) ->  Rational.is_positive r &&  not(Rational.is_zero r)
    | (_,_)-> failwith "evaluate : should not happen" 
	

let var_exists l = 
  List.exists (
    fun el -> match el with
    | OpNum(Times,_,Ref(t)) -> true
    | Ref(w) -> true
    | _ -> false
  ) l

(*check if a given variable s appears in (num)expression exn*)
let rec belongs_to_expn expn s = match expn with
  | OpNum(op,opn1,opn2) -> belongs_to_expn opn1 s|| belongs_to_expn opn2 s
  | Ref(t) when t=s -> true
  | _ -> false

(*check if a given variable s appears in expression exb*)
let rec belongs_to_expb expb s = 
  match expb with 
    |Term (ExprNum(op,expn1,expn2)) -> (belongs_to_expn expn1 s||  belongs_to_expn expn2 s)
    |Term(_)-> false
    |OpBoolB(op,expb1,expb2) -> belongs_to_expb expb2 s|| belongs_to_expb expb2 s
    |Neg(expb2)-> belongs_to_expb expb2 s

(*Substitute id (variable) by expid an expression, in the expression expn0*)
let rec substitute_in_expn expn0 id expid = 
  match expn0 with
    | OpNum(op,opn1,opn2) -> OpNum(op,substitute_in_expn opn1 id expid,
				   substitute_in_expn opn2 id expid)
    | Ref(t) when t=id-> expid 
    | _ as c -> c

(*The same, but within a numerical expression*)
let rec substitute_in_expb expb0 id expid = 
  match expb0 with 
    |Term (ExprNum(op,expn1,expn2)) -> Term (ExprNum(op,substitute_in_expn expn1 id expid,
						     substitute_in_expn expn2 id expid))
    |Term(_) as c -> c	
    |OpBoolB(op,expb1,expb2) -> OpBoolB(op,substitute_in_expb expb1 id expid,
					substitute_in_expb expb2 id expid) 
    |Neg(expb2) ->  Neg(substitute_in_expb expb2 id expid)

(*all vars in expressions*)


let rec vars_in_expn expn0 = 
  match expn0 with
    | OpNum(op,opn1,opn2) -> Sette.union (vars_in_expn opn1) (vars_in_expn opn2)
    | Ref(t) -> Sette.singleton t 
    | _ as c -> Sette.empty

let rec vars_in_expb expb = match expb with
    |Term (ExprNum(op,exp1,exp2)) ->  Sette.union 
       (vars_in_expn exp1) (vars_in_expn exp2)
    | Term(_) -> Sette.empty
    |OpBoolB(op,exp1,exp2) -> 
       Sette.union (vars_in_expb exp1) (vars_in_expb exp2)
    |Neg(expb2) ->  vars_in_expb expb2
       

    

let pp_print_expr_bool_term lvars fmt (et:term) = 
  match et with
    | ExprNum(opcompare,exp1,exp2) -> (
	let expminus = OpNum(Minus,exp1,exp2) in 
	let (l1,l2) = pos_and_neg expminus in
	let (l1b,l2b) = simplify_l l1 l2 lvars in 
	let t1 = List.length l1b and t2 = List.length l2b in
	  if (t1+t2 >= 2) || (var_exists l1b) || (var_exists l2b)  then begin
	    if t1=0 then pp_print_expr_num lvars fmt (CNat(0)) else
	    print_list_with_sep fmt (pp_print_expr_num lvars) l1b " + ";
	    fprintf fmt " %s " (opcompare2string opcompare) ;
	    if t2=0 then pp_print_expr_num lvars fmt (CNat(0)) else
	      print_list_with_sep fmt (pp_print_expr_num lvars) l2b " + ";
	  end
	  else begin
	    let b = ref false in 
	      if t1+t2=0 then b:= evaluate opcompare (CNat(0))
	      else if t1 = 0 
	      then b:= evaluate (neg_opcompare opcompare) (List.hd l2b) 
	      else b:= evaluate opcompare (List.hd l1b);
	      if !b then fprintf fmt "true" else fprintf fmt "false"
	  end
      )
    | Stateis(w) -> fprintf fmt "state = %s" (Symbol.to_string w)
    | VBool(w) -> fprintf fmt "%s" (Symbol.to_string w)
    | CBool(bb) -> if bb then fprintf fmt "true" else fprintf fmt "false"



let rec eval_true_or_false  = function 
  |  Term(t) ->  (
       match t with 
	 | CBool(true) -> (true,false)
	 | CBool(false) -> (false,true)
	 | Stateis(_) -> (false,false)
	 | VBool(_) -> (false,false)
	 | ExprNum(opcompare,exp1,exp2) -> ( (*cf printing*)
	     (*case  2<=42*)
	     let expminus = OpNum(Minus,exp1,exp2) in 
	     let lvars = find_lvars_num expminus in 
	     let (l1,l2) = pos_and_neg expminus in
	     let (l1b,l2b) = simplify_l l1 l2 lvars in 
	     let t1 = List.length l1b and t2 = List.length l2b in
	       if (t1+t2 >= 2) || (var_exists l1b) || (var_exists l2b)  
	       then (false,false)
	       else begin
		 let b = ref false in 
		   if t1+t2=0 then b:= evaluate opcompare (CNat(0))
		   else if t1 = 0 
		   then b:= evaluate (neg_opcompare opcompare) (List.hd l2b) 
		   else b:= evaluate opcompare (List.hd l1b);
		   (!b,not(!b))
	       end
	   )
     )
  |  OpBoolB(op,eb1,eb2) -> (
       let (un,deux)=eval_true_or_false eb1 and (unb,deuxb)=eval_true_or_false eb2 in 
	 match op with 
	   | Or ->  (un || deux ,deux && deuxb)
	   | And ->  (un && unb,deux || deuxb)
	   | Imply -> failwith "eval true, should not happen";
     )
  |  Neg(eb) -> let (b1,b2) = eval_true_or_false eb in 
       if b1 then (false,true) else if b2 then (true,false) else (false,false)



(*negate a given boolean formula*)
let rec neg_of_bool_expr bexp = match bexp with 
  | Term(ExprNum(op,en1,en2)) when op=NotEq-> Term(ExprNum(neg_opcompare op,en1,en2))
  | Term(ExprNum(op,en1,en2)) -> Term(ExprNum(neg_opcompare op,en1,en2))
  | Term(CBool(b)) -> Term(CBool(not b))
  | Term(VBool(v)) as t ->  (printf "ici@."; Neg(t))
  | Term(t) ->  failwith "negation de stateis.."
  | Neg(bexp1) -> bexp1
  | OpBoolB(op,bexp1,bexp2) when op = And -> 
      OpBoolB(Or,neg_of_bool_expr bexp1,neg_of_bool_expr bexp2)     
  | OpBoolB(op,bexp1,bexp2) when op = Or -> 
      OpBoolB(And,neg_of_bool_expr bexp1,neg_of_bool_expr bexp2)  
  | OpBoolB(op,bexp1,bexp2) when op = Imply -> 
      OpBoolB(And,bexp1,neg_of_bool_expr bexp2)  


(*simplify and throw away "or" in bexp*)
let rec enleve_les_ou  bexp = 
  let (t,f) = eval_true_or_false bexp in 
    if t then [Term(CBool(true))] else if f then [Term(CBool(false))]
  else match bexp with
  | Term(ExprNum(op,en1,en2)) when op = NotEq-> 
      Term(ExprNum(Less,en1,en2))::[Term(ExprNum(Sup,en1,en2))]
  | Term(t) as w -> [w]
  | OpBoolB(op,exp1,exp2) when op = And -> 
      begin
	match (exp1,exp2) with
	  | (Term(CBool(false)),_) -> [Term(CBool(false))]
	  | (Term(CBool(true)),_) -> enleve_les_ou exp2 
	  | (_,Term(CBool(false))) -> [Term(CBool(false))]
	  | (_,Term(CBool(true))) -> enleve_les_ou exp1
	  | _,_ -> (
	      let l1 = enleve_les_ou exp1 and l2 = enleve_les_ou exp2 in 
	      let res = ref [] in 
		List.iter (
		  function oneexp -> 
		    let aux = ref [] in
		      List.iter ( 
			fun one_el -> 
			  aux:=(OpBoolB(And,oneexp,one_el))::!aux;
		      )l2;
		      res:= !aux::!res
		) l1;
		List.flatten !res)
      end
  | OpBoolB(op,exp1,exp2) when op = Or -> begin
      match (exp1,exp2) with
	| (Term(CBool(false)),_) -> enleve_les_ou exp2 
	| (Term(CBool(true)),_) -> [Term(CBool(true))]
	| (_,Term(CBool(false))) -> enleve_les_ou exp2 
	| (_,Term(CBool(true))) -> [Term(CBool(true))]
	| (_,_) -> (
	    let l1 = enleve_les_ou exp1 and l2 = enleve_les_ou exp2 in l1@l2 )
    end
  | Neg(be) as t -> begin
      match be with 
	| Term(VBool(vp)) -> [t]
	| _ -> 
	    let nb = neg_of_bool_expr be in
	      enleve_les_ou nb
    end


(*Disjunctive normal for + simplication*)
let rec fnd_exprbool_aux bexp = 
    match bexp with
      | Term(CBool(true)) -> Some([])
      | Term(CBool(false)) -> None
      | _  -> (
	  let baux = enleve_les_ou bexp in 
	    (*filtrage des false*)
	  let (_,ll2) = List.partition
	    (
	      fun oneel -> match oneel with
		| Term(exp) -> 
		    let lvars = find_lvars_bool oneel in
		    ignore (Format.flush_str_formatter ());
		    pp_print_expr_bool_term lvars str_formatter exp ;
		    let s = Format.flush_str_formatter () in
		      s="false"
		| _ -> false
		
	    )
	    baux in
	    if List.length ll2 = 0 then None else Some(baux)
	)

let fnd_exprbool bexp = 
  let resu = fnd_exprbool_aux bexp in 
    match resu with 
      | None -> None
      | Some(onelist) -> 
	  if List.exists (fun el -> el=Term(CBool(false))) onelist 
	  then None
	  else 
	    let (_,res) = 
	      List.partition (fun el -> el=Term(CBool(true))) onelist in 
	      if (List.length res=0) then Some([]) else Some(res)

(**************print a boolean expression******************)

(*Aux function*)
let rec pp_print_expr_bool_aux lvars fmt bexp = 
  match bexp with
      Term(t) ->  pp_print_expr_bool_term lvars fmt t
    | Neg(bexp1) -> 
	(match bexp1 with
	   | Term(VBool(vb)) -> fprintf fmt "not(%s)" (Symbol.to_string vb)
	   | _ -> 
	       let bprime = neg_of_bool_expr bexp1 in
		 pp_print_expr_bool_aux lvars fmt bprime
	)
    | OpBoolB(op,bexp1,bexp2) ->  begin
	fprintf fmt "(";
	pp_print_expr_bool_aux lvars fmt bexp1;
	fprintf fmt ") %s (" (opboolB2string op);
	pp_print_expr_bool_aux lvars fmt bexp2;
	fprintf fmt ")";
      end
	
(*aux function*)
let pexp fmt bexp = match bexp with
  | Term(CBool(true)) -> fprintf fmt "true"
  | Term(CBool(false)) -> fprintf fmt "false"
  | Term(VBool(symb)) -> fprintf fmt "%s" (Symbol.to_string symb)
  | Neg(Term(VBool(symb))) -> fprintf fmt "not(%s)" (Symbol.to_string symb)
  | _ -> (
      let lvars = find_lvars_bool bexp in 
      let lfnd= fnd_exprbool bexp in begin
	  match lfnd with
	    | None -> fprintf fmt "false"
	    | Some(l) -> (
		let (_,lbis) = List.partition (
		  fun oneexp -> oneexp = Term(CBool(true))
		    
		) l in 
		  if (List.length lbis = 1) then pp_print_expr_bool_aux lvars fmt (List.hd lbis)
		  else
		    let x=ref (List.length lbis) in 
		      List.iter ( fun el ->
				    pp_print_expr_bool_aux lvars fmt el ;
				    if !x>1 then fprintf fmt "||" ; x:=!x-1) lbis;)
	end
    )

(**************************** INTERFACE**************************)

(*Main functions for printing expressions*)
let print_exprbool = pexp

let pnum fmt nexp = 
  match nexp with 
    | Indet(s) -> fprintf fmt "?"
    | _ ->   let lvars = find_lvars_num nexp in 
    pp_print_expr_num lvars fmt nexp

let print_exprnum = pnum

let exprbool2string exp = 
  ignore (Format.flush_str_formatter ());
  print_exprbool str_formatter exp;
  Format.flush_str_formatter ()

let exprnum2string exp = 
  ignore (Format.flush_str_formatter ());
  print_exprnum str_formatter exp;
  Format.flush_str_formatter ()

