(*fastDummyprinter, Copyright (C) 2012 Laure Gonnord*)
(*This file was part of Aspic Distribution, (C) Laure Gonnord 2007*)
(*  *)
(* This library is free software; you can redistribute it and/or *)
(* modify it under the terms of the GNU Lesser General Public *)
(* License as published by the Free Software Foundation; either *)
(* version 2.1 of the License, or (at your option) any later version. *)
(*  *)
(* This library is distributed in the hope that it will be useful, *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)
(*  *)
(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

(*Lexical Analysis for Fast files*)
{
open ParserFast
open Lexing

let pow10 n = 
  let res = ref 1 in
  for i=0 to n-1 do
    res := !res * 10
  done;
  !res

(*c Keywords *)
let keywords = Hashtbl.create 42
let _ =
  Array.iter
    (fun (keyword,token) -> Hashtbl.add keywords keyword token)
    [| 
	(*for fast*)
      ("model", TK_MODEL);
      ("var",TK_VAR);
      ("states",TK_STATES);
      ("transition",TK_TRANS);
      ("from",TK_FROM);
      ("to",TK_TO);
      ("guard",TK_GUARD);
      ("action",TK_ACTION);
      ("exists",TK_EXISTS);
      ("pre",TK_PRE);
      ("post",TK_POST);
      ("false",TK_FALSE);
      ("true", TK_TRUE);
      ("strategy", TK_STRATEGY);
      ("Region",TK_REGION);
      ("init", TK_REGINIT);
      ("bad", TK_REGBAD);
      ("state", TK_STATE); 
      ("Transitions",TK_STRATTRANS);
      ("boolean",TK_BOOLEAN);
      ("True", TK_STRATTRUE);
      ("False", TK_STRATFALSE);
      ("eqSet", TK_STRATSUBSET);
      ("subSet", TK_STRATEQSET);
      ("isEmpty", TK_STRATISEMPTY);
      ("setMaxState",TK_SETMAXSTATE);
      ("setMaxAcc",TK_SETMAXACC);
      ("print",TK_PRINT);
      ("if",TK_IF);
      ("endif",TK_ENDIF);
      ("then",TK_THEN);
      ("else",TK_ELSE);
      ("and",TK_AND);
      ("or",TK_OR);
      ("parameters",TK_PARAM);

    |]
} 


rule comment = parse                                                         
    [^'*' '/' '\n']* "*/" {()}                                                  
   |[^'/' '*' '\n']* "/*" {comment lexbuf; comment lexbuf}                      
   |[^'/' '*' '\n']* '/'  {comment lexbuf}                                      
   |[^'*' '/' '\n']* '*'  {comment lexbuf}                                      
   |[^'*' '/' '\n']* '\n'        {Parse.lex_eol lexbuf;comment lexbuf}
and  lexBac = parse
  | [' ' '\t'] +  { lexBac lexbuf }
  | ['\n'] { Parse.lex_eol lexbuf; lexBac lexbuf }
  | "//@" { TK_AT }
  | "//" [^'\n']* '\n'?  { lexBac lexbuf}
  | "/*" { comment lexbuf ; lexBac lexbuf}
  | eof    { TK_EOF }

(* Dlimiteurs *)
| ","   { TK_COMMA }
| ";"   { TK_SEMI }
| '''   { TK_PRIME }
| "("   { TK_LPAREN }
| ")"   { TK_RPAREN }
| "{"   { TK_LBRACE }
| "}"   { TK_RBRACE }
| "["   { TK_LBRACKET }
| "]"   { TK_RBRACKET }
| ":"   { TK_COLON }
| "\""   { TK_DOUBLEQUOTE }

(* Oprations arithmtiques *)
| "+"   { TK_PLUS }
| "-"   { TK_MINUS }
| "*"   { TK_MUL }

(* Oprations mixtes *)
| ">"   { TK_GREATER }
| "<"   { TK_LESS }
| ">="  { TK_GREATEREQUAL }
| "<="  { TK_LESSEQUAL }
| "="   { TK_EQUAL }
| "!="   { TK_NOTEQUAL }

(*assign*)
| ":="  { TK_ASSIGN } 

(*bool ops*)
| "&&" {TK_ANDAND}
| "||" {TK_OROR}
| "!" {TK_EXCLA}
| "=>" {TK_IMPLY}
| "<=>" {TK_EQUIV}

(*indet*)
| "?" {TK_INDET}

(* Numbers *)
| ['0'-'9']+'/'['0'-'9']+
  {
    let str = lexeme lexbuf in
    let index = String.index str '/' in
    let num = int_of_string (String.sub str 0 index) in
    let den = int_of_string (String.sub str (index+1) ((String.length str)-(index+1))) in
    let r = Rational.make num den in
    TK_RAT(r)
  }
| ['0'-'9']* ('.' ['0'-'9']+) (['e' 'E'] ['+' '-']? ['0'-'9']+)?
  { 
    let str = lexeme lexbuf in
    begin
      let num = float_of_string str in
      if num <> 0.0 && abs_float (log10 num) > 8.0 then begin
	Format.eprintf "Lexical analysis: floating-point constant %s of too small or too big magnitude (maximum: abs (log10 flt) <= 8)@." str;
	raise Parse.Lex_error
      end;
    end;
    let str = String.uppercase str in
    let (eindex, exponent) = 
      try 
	let eindex = String.index str 'E' in
	let eindex1 =
	  if str.[eindex+1] = '+' then eindex+2 else eindex+1
	in
	let substr = String.sub str eindex1 (String.length str - eindex1) in
	(eindex, int_of_string substr)
      with Not_found ->
	(String.length str, 0)
    in
    let dotindex = String.index str '.' in
    let nbaftercomma = eindex - (dotindex+1) in
    let num1 = int_of_string (String.sub str 0 dotindex) in
    let num2 = int_of_string (String.sub str (dotindex+1) nbaftercomma) in
    let den = pow10 nbaftercomma in
    let r = Rational.make (num1*den+num2) den in
    let nr = 
      if exponent = 0 then 
	r
      else if exponent > 0 then { 
	Rational.num = r.Rational.num * (pow10 exponent); 
	Rational.den = r.Rational.den 
      }
      else (* exponent < 0 *) { 
	Rational.num = r.Rational.num;
	Rational.den = r.Rational.den  * (pow10 (-exponent)); 
      }
    in
    TK_RAT(nr)
  }
| ['0'-'9']+ 
    { TK_NUM(int_of_string (lexeme lexbuf)) }

(* Identificateurs *)
| ['A'-'Z' 'a'-'z' '_'] ( ['_' '.' 'A'-'Z' 'a'-'z' '0'-'9'] ) *
    { let id = lexeme lexbuf in
      try Hashtbl.find keywords id 
      with Not_found -> TK_ID (Symbol.add id) }
| ['A'-'Z' 'a'-'z' '_'] ( ['_' '.' 'A'-'Z' 'a'-'z' '0'-'9' '!'] ) *
    { let id = lexeme lexbuf in
      try Hashtbl.find keywords id 
      with Not_found -> TK_IDEXCLA (Symbol.add id) }
| _         
    { raise Parse.Lex_error }
